
--Copyright (c) 1993 Ohio Board of Regents and the University of
--Cincinnati.  All Rights Reserved.

--author: philip a. wilsey
--phil.wilsey@uc.edu


library compose;
use compose.compose.all;
use std.textio.all;

entity pe is

    generic (
    	id : string := "Unnamed PE";
	intid : integer;
    	debug_control : debug := none);
	
    port (
	read_mm : inout resolved_bit;
	write_mm : out resolved_bit;
	mar : out register_type;
	mdr : inout register_type;
	mfc : in bit;

	memory_initialize : in bit;
	memory_initialized : out bit;
	
        all_listen : in bit;
	opcode_to_execute : in bit;
	
	pc_to_mar : in bit;
	increment_pc : in bit;
	mdr_to_ir : in bit;
	
	s0  :  in bit;
	s1  :  in bit;
	s2  :  in bit;
	s3  :  in bit;
	s4  :  in bit;
	s5  :  in bit;
	s6  :  in bit;
	s7  :  in bit;
	s8  :  in bit);
    

    type register_file_type is array (0 to 63) of register_type;
    signal register_file : register_file_type;
    alias registers : register_file_type is register_file;
    
    signal instr_reg : register_type;
    signal work      : register_type;
    signal condition_codes : bit_vector(2 downto 0); --EQ, LT, GT

    constant pc : integer := 63;
    constant c_init : integer := 0;    -- address of first instr. in memory
    constant peid : string := id & stringof(intid); -- pied piper of hamelin
end;


architecture behavior of pe is

signal BC : bit := '0';
alias opcode : bit_vector (8 downto 0) is instr_reg(30 downto 22);
alias op0 : bit is instr_reg(30);
alias op1 : bit is instr_reg(29);
alias op2 : bit is instr_reg(28);
alias op3 : bit is instr_reg(27);
alias op4 : bit is instr_reg(26);
alias op5 : bit is instr_reg(25);
alias op6 : bit is instr_reg(24);
alias op7 : bit is instr_reg(23);
alias op8 : bit is instr_reg(22);
alias mult : bit is instr_reg(21);
alias L : bit is instr_reg(31);
alias br_conds : bit_vector (2 downto 0) is instr_reg(20 downto 18);
alias a_reg : bit_vector (5 downto 0) is instr_reg(17 downto 12);
alias b_reg : bit_vector (5 downto 0) is instr_reg( 11 downto  6);
alias c_reg : bit_vector (5 downto 0) is instr_reg( 5 downto  0);
alias mdr_opcode : bit_vector (8 downto 0) is mdr(30 downto 22);
alias mdr_mult : bit is mdr(21);
alias mdr_br_conds : bit_vector (2 downto 0) is mdr(20 downto 18);
alias mdr_a_reg : bit_vector (5 downto 0) is mdr(17 downto 12);
alias mdr_b_reg : bit_vector (5 downto 0) is mdr(11 downto  6);
alias mdr_c_reg : bit_vector (5 downto 0) is mdr( 5 downto  0);

signal a_reg_int : integer;
signal b_reg_int : integer;
signal c_reg_int : integer;

--these functions require a decending range

function logical_l(src : bit_vector; d : integer) return bit_vector is

begin

    assert src'length > d
        report "logical_l: shift distance exceeds size of input vector."
	severity failure;
	
    if (d = 0) then return src;
               else return logical_l(src(src'left-1 downto src'right) & '0',d-1);
    end if;

end logical_l;

function logical_r(src : bit_vector; d : integer) return bit_vector is

begin

    assert src'length > d
        report "logical_r: shift distance exceeds size of input vector."
	severity failure;
	
    if (d = 0) then return src;
               else return logical_r('0' & src(src'left downto src'right+1),d-1);
    end if;

end logical_r;

function arith_r(src : bit_vector; d : integer) return bit_vector is

begin

    assert src'length > d
        report "arith_r: shift distance exceeds size of input vector."
	severity failure;
	
    if (d = 0) then return src;
               else return arith_r(src(src'left) & src(src'left downto src'right-1),d-1);
    end if;

end arith_r;

function circle_l(src : bit_vector; d : integer) return bit_vector is

begin

    assert src'length > d
        report "circle_l: shift distance exceeds size of input vector."
	severity failure;
	
    if (d = 0) then return src;
               else return circle_l(src(src'left-1 downto src'right) & src(src'left),d-1);
    end if;

end circle_l;

function circle_r(src : bit_vector; d : integer) return bit_vector is

begin

    assert src'length > d
        report "circle_r: shift distance exceeds size of input vector."
	severity failure;
	
    if (d = 0) then return src;
               else return circle_r(src(src'right) & src(src'left downto src'right+1) & src(src'left),d-1);
    end if;

end circle_r;

procedure log_opcode (debug_cntrl : debug; opcode : string) is

begin
    if (debug_cntrl = trace_opcodes or
        debug_cntrl = trace_pe or
        debug_cntrl = everything) then
            print_debug(peid, debug_cntrl, "Opcode: " & opcode);
    end if;
end log_opcode;

begin


startup : process

    variable bv_line : line;
    variable mar_bv : bit_vector (31 downto 0);
    variable mdr_bv : bit_vector (31 downto 0);
    file test_in : TEXT is in "infile."&stringof(intid);

begin
    wait until memory_initialize = '1';
    while not endfile(test_in) loop

	readline(test_in, bv_line);

	if bv_line'length = 64 then
	    read(bv_line, mar_bv);
	    mar <= mar_bv;
	else 
	    assert false
	        report "Error in memory initialization input file."
		severity failure;
	end if;

	if bv_line'length = 32 then
	    read(bv_line, mdr_bv);
	    mdr <= mdr_bv;
	else 
	    assert false
	        report "Error in memory initialization input file."
		severity failure;
	end if;
	
	write_mm <= '1';
	if (debug_control = control_loop or
	    debug_control = memory_initialization or
	    debug_control = everything) then
	        print_debug(peid, debug_control, 
		            "Initializing all PE memories: MM" &
			    stringof(intof(mar_bv)) & "] = " &
			    stringof(mdr_bv));
	end if;

	wait until mfc = '1';
	write_mm <= '0';
        WAIT FOR 0 ns;
    end loop;

    mar <= (others => '0');  -- disconnect.  these should be null, but i'm to lazy to update the vhdl code
    mdr <= (others => '0');  -- disconnect
    memory_initialized <= '1';
    if (debug_control = control_loop or
	debug_control = memory_initialization or
	debug_control = everything) then
	    print_debug(peid,debug_control,"PE memory initialization complete.");
    end if;

    wait;

end process;


run : process begin
    wait on all_listen, pc_to_mar, increment_pc, mdr_to_ir,read_mm,
    opcode_to_execute, s0, s1, s2, s3, s4, s5, s6, s7, s8;

    if all_listen'event and all_listen = '1' then
	    register_file(0) <= X"00000000";
	    register_file(pc) <= bitvectorof(c_init);
    end if;


    if pc_to_mar'event and pc_to_mar = '1' then
	    mar <= registers(pc);
	    if (debug_control = trace_pe or
	        debug_control = everything) then
	        print_debug(peid, debug_control, "Moving pc to mar: " &
		                              stringof(intof(registers(pc))));
	    end if;
    end if;

    if increment_pc'event and increment_pc = '1' then
	    registers(pc) <= bitvectorof(intof(registers(pc)) + 4);
	    if (debug_control = trace_pe or
	        debug_control = everything) then
	        print_debug(peid, debug_control, "Incrementing pc to: " &
		                              stringof(intof(registers(pc)) + 4));
	    end if;
    end if;

    if mdr_to_ir'event and mdr_to_ir = '1' then
	    instr_reg <= mdr;
	    a_reg_int <= intof('0' & mdr_a_reg);
	    b_reg_int <= intof('0' & mdr_b_reg);
	    c_reg_int <= intof('0' & mdr_c_reg);
	    if (debug_control = trace_pe_instrs or
	        debug_control = trace_pe or
	        debug_control = everything) then
	        print_debug(peid, debug_control, "New IR:" &
                               " Opcode: " & stringof(mdr_opcode) &
                               " Mult: " & stringof(mdr_mult) &
                               " br conditions: " & stringof(mdr_br_conds) &
                               " A reg: " & stringof(intof('0' & mdr_a_reg)) &
                               " B reg: " & stringof(intof('0' & mdr_b_reg)) &
                               " C reg: " & stringof(intof('0' & mdr_c_reg)));

	    end if;
    end if;

    if s7'event and s7 = '1' and op7 = '1' then
	    if L = '1' then
	      mar <= bitvectorof(intof(registers(b_reg_int)));
	      read_mm <= '1';
	      log_opcode(debug_control, "load: initiating mm read.");

	      wait until mfc = '1';
	      registers(c_reg_int) <= mdr;
	      read_mm <= '0';
	      log_opcode(debug_control,
	       	         "load: R[" & stringof(a_reg_int) & "] = MM[" &
		         stringof(intof(bitvectorof(intof(registers(b_reg_int))))) & "] = " & stringof(mdr));
	    else
	      if BC = '1' then
		registers(c_reg_int) <= registers(pc);
	      end if;
	      if (br_conds = b"000" or br_conds = b"100" ) then
		registers(b_reg_int) <= registers(a_reg_int);
	      end if;
	    end if;
     end if;


    if s0'event and s0 = '1' then
	    if L = '0' and op0 = '0' then
		work <= registers(a_reg_int);
		log_opcode(debug_control,
			   "work = " & stringof(intof(registers(a_reg_int))));
	    elsif L = '0' and op0 = '1' then
		work <= not registers(a_reg_int);
		log_opcode(debug_control,
			   "work = " & stringof(intof(not registers(a_reg_int))));
	    elsif L = '1' and op0 = '0' then
		work <= b"00000000000000000000000000" & a_reg;
		log_opcode(debug_control,
		   "work = " & stringof((intof(b"00000000000000000000000000" & a_reg))));
	    elsif L = '1' and op0 = '1' then
		work <= b"11111111111111111111111111" & a_reg;
		log_opcode(debug_control,
		   "work = " & stringof((intof(b"11111111111111111111111111" & a_reg))));
	    end if;
    end if;

    if s8'event and s8 = '1' and op8 = '1' then
	    if L = '1' then
	      mar <= bitvectorof(intof(registers(b_reg_int)));
	      mdr <= registers(c_reg_int);
	      write_mm <= '1';
	      log_opcode(debug_control,
 	         "store: MM[ " & stringof(intof(registers(b_reg_int))) &
	         "] = R[" & stringof(a_reg_int) & "] = " &
	         stringof(registers(a_reg_int)));
	
	      wait until mfc = '1';
	      write_mm <= '0';
	      log_opcode(debug_control, "store: completed mm write.");
	    else
	      if (BC = '1') then
		registers(pc) <= registers(a_reg_int);
	      end if;
	      if (br_conds = b"000" or br_conds = b"100") then
		registers(b_reg_int) <= registers(a_reg_int);
	      end if;
	    end if;
   end if;

    if ((s1'event and s1 = '1' and op1 = '1') or (s5'event and s5 = '1' and
         op5 = '1')) then
	    registers(b_reg_int) <= not registers(b_reg_int);
	    case br_conds is
		when "000" =>
		when "001" => 
		   if (intof(not registers(b_reg_int)) = 0) then
		      BC <= '1';
		   end if;
	        when "010" =>
		   if (intof(not registers(b_reg_int)) < 0) then
		      BC <= '1';
		   end if;
		when "011" =>
		   if (intof(not registers(b_reg_int)) > 0) then
		      BC <= '1';
		   end if;
		when "100" =>
		when "101" =>
		   if (intof(not registers(b_reg_int)) = 0) then
		      BC <= '0';
		   end if;
		when "110" =>
		   if (intof(not registers(b_reg_int)) < 0) then
		      BC <= '0';
		   end if;
		when "111" =>
		   if (intof(not registers(b_reg_int)) > 0) then
		      BC <= '0';
		   end if;
		end case;
   end if;


   if (s2'event and s2 = '1' and op2 = '1') then
	    registers(b_reg_int) <= bitvectorof(intof(registers(b_reg_int))+1);
	    log_opcode(debug_control,
		       "incb: register " & stringof(b_reg_int) &
		       " original contents " & stringof(registers(b_reg_int)) &
		       " new contents "
		       & stringof(bitvectorof(intof(registers(b_reg_int))+1)));


   end if;


    if (s3'event and s3 = '1' and op3 = '1') then
	    if (mult = '0' or registers(c_reg_int)(0) = '1') then

		registers(b_reg_int) <= bitvectorof(intof(work) +
		                                    intof(registers(b_reg_int)));
		log_opcode(debug_control,
	           "add: reg(b) = reg(b) + reg(a). b: " &
		   stringof(intof(registers(b_reg_int))) & " a: " & stringof(intof(work)));
		log_opcode(debug_control,
		   " result " & stringof((intof(work))
		   + (intof(registers(b_reg_int)))));

	    else

		log_opcode(debug_control,
		           "add: reg(b) = reg(b) + reg(a). b: " &
			   stringof(b_reg_int) & " a: " & stringof(a_reg_int) &
			   " NOT EXECUTED");
	    end if;

   end if;


    if (s4'event and s4 = '1' and op4 = '1') then
	    registers(b_reg_int) <= work nand registers(b_reg_int);
	    log_opcode(debug_control,
		       "nand: reg(b) = work nand reg(b). a: " &
		        stringof(intof(work)) & " b: " & stringof(intof(registers(b_reg_int))) &
		       " result: " &
		       stringof(work nand registers(b_reg_int)));
   end if;	




    if (s6'event and s6 = '1' and op6 = '1' ) then
	    if (mult = '1') then
	        if (opcode(opcode'right) = '0') then

		    registers(c_reg_int) <= logical_r(registers(c_reg_int),1);
		    log_opcode(debug_control, "shft: logical right 1-bit " &
		               "input: " & stringof(registers(c_reg_int)) &
		               " result: " & stringof(logical_r(registers(c_reg_int),1)));

		else

		    registers(c_reg_int) <= logical_l(registers(c_reg_int),1);
		    log_opcode(debug_control, "shft: logical left 1-bit " &
		               "input: " & stringof(registers(c_reg_int)) &
			       " result: " & stringof(logical_l(registers(c_reg_int),1)));

		end if;
	    elsif (registers(b_reg_int)(7) = '1') then

		registers(c_reg_int) <= arith_r(registers(c_reg_int),intof('0' & registers(b_reg_int)(4 downto 0)));
		log_opcode(debug_control, 
          "shft: arith right of " & stringof(intof('0' & registers(b_reg_int)(4 downto 0))) &
           "bits. " & "input: " & stringof(registers(c_reg_int)) & " result: " & 
           stringof(arith_r(registers(c_reg_int),intof('0' & registers(b_reg_int)(4 downto 0)))));

	    elsif (registers(b_reg_int)(6) = '0' and registers(b_reg_int)(5) = '0') then

		registers(c_reg_int) <= logical_r(registers(c_reg_int),intof('0' & registers(b_reg_int)(4 downto 0)));
		log_opcode(debug_control, 
          "shft: logical right of " & stringof(intof('0' & registers(b_reg_int)(4 downto 0))) &
          "bits. " & "input: " & stringof(registers(b_reg_int)) & " result: " & 
          stringof(logical_r(registers(b_reg_int),intof('0' & registers(b_reg_int)(4 downto 0)))));

	    elsif (registers(b_reg_int)(6) = '0' and registers(b_reg_int)(5) = '1') then

	registers(c_reg_int) <= circle_r(registers(c_reg_int),intof('0' & registers(b_reg_int)(4 downto 0)));
	log_opcode(debug_control, 
           "shft: circular right of " & stringof(intof('0' & registers(b_reg_int)(4 downto 0))) &
           "bits. " & "input: " & stringof(registers(b_reg_int)) & " result: " & 
	   stringof(circle_r(registers(b_reg_int),intof('0' & registers(b_reg_int)(4 downto 0)))));

	    elsif (registers(b_reg_int)(6) = '1' and registers(b_reg_int)(5) = '0') then

		registers(c_reg_int) <= logical_l(registers(c_reg_int),intof('0' & registers(b_reg_int)(4 downto 0)));
		log_opcode(debug_control, 
           "shft: logical left of " & stringof(intof('0' & registers(b_reg_int)(4 downto 0))) &
           "bits. " & "input: " & stringof(registers(b_reg_int)) & " result: " & 
		           stringof(logical_l(registers(b_reg_int),intof('0' & registers(b_reg_int)(4 downto 0)))));

	    elsif (registers(b_reg_int)(6) = '1' and registers(b_reg_int)(5) = '1') then

		registers(c_reg_int) <= circle_l(registers(c_reg_int),intof('0' & registers(b_reg_int)(4 downto 0)));
		log_opcode(debug_control, 
		           "shft: circular left of " & stringof(intof('0' & registers(b_reg_int)(4 downto 0))) &
		           "bits. " & "input: " & stringof(registers(b_reg_int)) & " result: " & 
		           stringof(circle_l(registers(b_reg_int),intof('0' & registers(b_reg_int)(4 downto 0)))));

	    end if;

   end if;
end process;


end behavior;




